import torch
import torch.utils.data as td
from typing import Optional, Dict, Union
from transformers import BatchEncoding
from argparse import Namespace
import numpy as np
import pandas as pd
import pytorch_quik.utils as pu

Tensor_Target = Union[str, np.ndarray]
Tensor_Data = Union[pd.DataFrame, torch.Tensor, BatchEncoding]


def make_TensorDataset(
    tens: Tensor_Data,
    labels: Tensor_Target,
    ttype: Optional[str] = "train",
    data_types: Optional[Dict[int, type]] = None,
    args: Optional[Namespace] = None,
) -> td.TensorDataset:
    """Will turn a set of data into tensors in a torch TensorDataset for use
    in a Neural Network. Also provides the option of saving the TensorDataset

    Args:
        tens (Union[torch.Tensor, BatchEncoding]): Either a torch.Tensor,
            or if from transformers, a BatchEncoding.
        labels (Union[str, np.ndarray]): Can either be a string of the label
            name found in tens, or the actual labels as an np.ndarray
        args (Namespace, optional): The argparse arguments for the job.
            Defaults to None.
        ttype (str, optional): The type of dataset (train, valid, test).
            Defaults to "train".
        data_types (Dict[int, type], optional): If the tensor data types
            need to be changed for space considerations. Defaults to None.

    Returns:
        td.TensorDataset: The final TensorDataset
    """
    if data_types is not None:
        for i, col in enumerate(tens):
            tens[col] = tens[col].astype(data_types[i])
    if isinstance(labels, str):
        ttar = tens.pop(labels)
    if isinstance(tens, BatchEncoding):
        tds = td.TensorDataset(
            tens["input_ids"], tens["attention_mask"], torch.LongTensor(labels)
        )
    else:
        tens = torch.tensor(tens.values)
        tens = tens.transpose(0, 1)
        ttar = torch.tensor(ttar.values)
        tds = td.TensorDataset(*tens, ttar)
    if args is not None:
        tds_id = pu.id_str(ttype, args)
        torch.save(tds, tds_id)
    return tds


def transform_TensorDataset(
    tds: td.TensorDataset,
    pop: Optional[str] = None,
    torch_types: Optional[Dict[int, torch.dtype]] = None,
) -> td.TensorDataset:
    """Transforms a torch.utils.data.TensorDataset by either popping out
    a tensor, changing the data types, or both.

    Args:
        tds (td.TensorDataset): The original TensorDataset
        pop (str, optional): The ordinal of the tensor to pop. Defaults to
            None.
        torch_types (Dict[int, torch.dtype], optional): The ordinal and
            new data type of each tensor. Defaults to None.

    Returns:
        td.TensorDataset: The final transformed TensorDataset
    """
    tl = list(tds.tensors)
    if pop is not None:
        del tl[pop]
    if torch_types is not None:
        tl = [tens.type(torch_types[i]) for i, tens in enumerate(tl)]
    return td.TensorDataset(*tl)



import time
from pathlib import Path
from typing import Optional
from argparse import Namespace


def sec_str(st: float) -> str:
    """Simple util that tells the amount of time for a codeset (in seconds)

    Args:
        st (time): The start time of the codeset

    Returns:
        str: time in seconds
    """
    return str(round(time.time() - st, 2)) + " seconds"


def row_str(dflen: int) -> str:
    """String wrapper for the million of rows in a dataframe

    Args:
        dflen (int): the length of a dataframe

    Returns:
        str: rows in millions
    """
    return str(round(dflen / 1000000, 1)) + "M rows"


def id_str(
    ftype: str,
    args: Namespace,
    epoch: Optional[str] = None,
    gpu: Optional[str] = None,
    suffix: Optional[str] = ".pt",
) -> str:
    """Method to determine an appropriate filename for tensors, models, and
    state_dicts

    Args:
        ftype (str): file type (state_dict, train, valid, test, preds, model)
        args (Namespace): the current set of arguments. requires date,
            optional bert_type, source
        epoch (Optional[str], optional): the current epoch to be saving.
            Defaults to None.
        gpu (Optional[str], optional): the current gpu (for preds). Defaults
            to None.
        suffix (Optional[str], optional): file extension. Defaults to ".pt".

    Returns:
        str: full filename for output object
    """
    lbls = "".join(map(str, getattr(args, "labels", "")))
    if ftype == "state_dict":
        epoch = "e" + str(epoch)
    elif ftype == "preds":
        suffix = ".csv"
    else:
        ftype = ftype + "_tensor"
    id_list = [ftype, lbls, epoch, str(args.data_date), gpu]
    id_str = "_".join(filter(None, id_list))
    path_list = [
        "data",
        getattr(args, "bert_type", None),
        getattr(args, "source", None),
        id_str,
    ]
    path_list = filter(None, path_list)
    filename = Path.cwd().joinpath(*path_list).with_suffix(suffix)
    filename.parent.mkdir(parents=True, exist_ok=True)
    return filename



import numpy as np
from torch import Tensor
from transformers import (
    BertTokenizer,
    RobertaTokenizer,
    BertForSequenceClassification,
    RobertaForSequenceClassification,
    BatchEncoding,
)
from typing import Optional, List, Union

Tokenizers = Union[BertTokenizer, RobertaTokenizer]
Models = Union[BertForSequenceClassification, RobertaForSequenceClassification]

BERT_MODELS = {
    "bert": {
        "model-type": "bert-base-uncased",
        "tokenizer": BertTokenizer,
        "model": BertForSequenceClassification,
    },
    "roberta": {
        "model-type": "roberta-base",
        "tokenizer": RobertaTokenizer,
        "model": RobertaForSequenceClassification,
    },
}

BATCH_ENCODE_KWARGS = {
    "add_special_tokens": True,
    "return_attention_mask": True,
    "pad_to_max_length": True,
    "max_length": 256,  # max is 512
    "truncation": True,
    "return_tensors": "pt",
}


def get_tokenizer(bert_type: Optional[str] = "bert") -> Tokenizers:
    """Get a BERT or RoBERTa tokenizer from transformers to create
    encodings.

    Args:
        bert_type (str, optional): The type of tokenizer ("bert" or "roberta").
        Defaults to "bert".

    Returns:
        Tokenizers: The tokenizer to be used by get_encodings.
    """
    bert_dict = BERT_MODELS[bert_type]
    tokenizer = bert_dict["tokenizer"].from_pretrained(
        bert_dict["model-type"], do_lower_case=True
    )
    return tokenizer


def get_encodings(
    array_list: List[np.ndarray], idx: int, bert_type: Optional[str] = "bert"
) -> List[BatchEncoding]:
    """Get BERT or RoBERTa encodings for a list of np.ndarrays for training,
    testing (and validation).

    Args:
        array_list (List[np.ndarray]): The list of training, test, (and
        validation) arrays.
        idx (int): The column index of the text in the original pd.DataFrame
        bert_type (str, optional): The type of bert model to be used. Defaults
        to "bert".

    Returns:
        List[BatchEncoding]: A list of training, test (and validation)
        encodings of text.
    """
    tknzr = get_tokenizer(bert_type)
    encode_list = [
        tknzr.batch_encode_plus(array[:, idx], **BATCH_ENCODE_KWARGS)
        for array in array_list
    ]
    return encode_list


def get_pretrained_model(
    labels: Tensor, bert_type: Optional[str] = "bert"
) -> Models:
    """Get a pretrained model from transformers.

    Args:
        labels (Tensor): A tensor of the labels in the dataset. This will
        affect the size of the network final layer
        bert_type (str, optional): The type of the model to be used ("bert" or
        "roberta"). Defaults to "bert".

    Returns:
        Models: The pretrained model to be used in training.
    """
    num_lbls = len(np.unique(labels))
    bert_dict = BERT_MODELS[bert_type]
    model = bert_dict["model"].from_pretrained(
        bert_dict["model-type"],
        num_labels=num_lbls,
        output_attentions=False,
        output_hidden_states=False,
    )
    return model



import pytorch_quik.ddp
import pytorch_quik.tensors
import pytorch_quik.transform
import pytorch_quik.utils

__all__ = ["ddp", "tensors", "transform", "utils"]
__version__ = "0.0.2"

try:
    import pytorch_quik.bert
    __all__.append("bert")
except ImportError:
    print("skipping bert functions")



#!/usr/bin/env python

from setuptools import setup, find_packages
from pathlib import Path

here = Path(__file__).parent
README = here.joinpath("README.md").read_text()

with open(here.joinpath("requirements.txt")) as f:
    all_reqs = f.read().split("\n")

install_requires = [x.strip() for x in all_reqs if "git+" not in x]

setup(
    name="pytorch-quik",
    version="0.0.2",
    description="functions to make working in pytorch quik-er",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/donchesworth/dask-quik",
    author="Don Chesworth",
    author_email="donald.chesworth@gmail.com",
    license="BSD",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(exclude=["tests"]),
    install_requires=install_requires,
)



import dask_quik as dq
import torch
import torch.multiprocessing as mp
import torch.distributed as dist
from typing import Callable
from argparse import Namespace
import os


def train(train_fn: Callable, args: Namespace):
    """Determine whether to train using DistributedDataParallel
    or a single training instance

    Args:
        train_fn (Callable): The train function for this particular
        Neural Network
        args (Namespace): The argparse Namespace for this script
    """
    if bool(dq.utils.gpus()):
        ddp_train(train_fn, args)
    else:
        train_fn(gpu=None, args=args)


def ddp_train(train_fn: Callable, args: Namespace):
    """Using multiprocessing, initiate a DistributedDataParallel
    training process

    Args:
        train_fn (Callable): The train function for this particular
        Neural Network
        args (Namespace): The argparse Namespace for this script
    """
    mp.spawn(train_fn, nprocs=args.gpus, args=(args,))


def setup(gpu: str, args: Namespace):
    """Setup of the distrubtion settings.

    Args:
        gpu (str): The current gpu
        args (Namespace): The argparse Namespace for this script
    """
    os.environ["MASTER_ADDR"] = "localhost"
    os.environ["MASTER_PORT"] = "12355"
    os.environ["NCCL_P2P_LEVEL"] = "0"
    rank = args.nr * args.gpus + gpu

    dist.init_process_group(
        backend="nccl",
        init_method="env://",
        world_size=args.world_size,
        rank=rank,
    )

    # Explicitly setting seed to make sure that models created in two processes
    # start from same random weights and biases.
    torch.manual_seed(42)


def cleanup():
    """Deletion of the DistributedDataParallel process group"""
    dist.destroy_process_group()



from typing import Optional, Union, Dict, List, Tuple
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from imblearn.over_sampling import RandomOverSampler


def create_dicts(
    categories: np.array, labels: np.array
) -> Tuple[Dict[int, int], Dict[int, int]]:
    """Turns two lists of ints into two dictionaries with a index key and
    a value from the list. These help with a quick relabelling.

    Args:
        categories (List[int]): The categorical variable numbers
        labels (List[int]): The label numbers

    Returns:
        Tuple[Dict[int, int], Dict[int, int]]: A category indexed dictionary
        and a category keys labels values dictionary.
    """
    catd = dict(zip(categories, range(0, len(categories))))
    lbld = dict(zip(categories, labels))
    return catd, lbld


def create_labels(
    df: pd.DataFrame,
    labels: Optional[List[int]] = [0, 0, 1, 2, 2],
    category: Optional[str] = "category",
) -> pd.DataFrame:
    """Turn categorical variables into labels to be predicted. For instance,
    if your data is 1-10, and you want to predict 5-10 is positive, and 0-4
    is negative, then you would provide the category column name, and a list
    of [0, 0, 0, 0, 1, 1, 1, 1, 1, 1]. This will label all rows as positive or
    negative.

    Args:
        df (pd.DataFrame): Original df to be labelled.
        category (str, optional): DataFrame column name containing the
            category. Defaults to 'category'.
        labels (List[int], optional): The labels to be assigned to each
            category. It must be the same length as the number of unique
            categories. Defaults to [0, 0, 1, 2, 2].

    Returns:
        pd.DataFrame: The original df with an additional column named
        label.
    """
    cats = np.sort(df[category].unique())
    labels = np.array(labels)
    print("Categories\t\t" + str(cats) + "\nNow match labels\t" + str(labels))
    cat_dict, label_dict = create_dicts(cats, labels)
    df["cat_index"] = df[category].replace(cat_dict)
    df["label"] = df[category].replace(label_dict)
    return df


def imbalanced_split(
    df: pd.DataFrame,
    label_name: Optional[str] = "label",
    test_size: Optional[int] = 0.1,
    valid_size: Optional[Union[int, None]] = None,
    random_state: Optional[int] = 42,
) -> Tuple[np.ndarray]:
    """When categories are imbalanced, create a balanced split of train,
    validation, and test sets.

    Args:
        df (pd.DataFrame): The original dataFrame to be split
            label_name (str, optional): the name of the
            column to be balanced. Defaults to "label".
            test_size (int, optional): The size of the test set.
            Defaults to 0.1.
        valid_size (Union[int, None], optional): The size of the
            validation set. Defaults to None.
        random_state (int, optional): The random state, or seed.
            Defaults to 42.

    Returns:
        Tuple[np.ndarray]: A tuple of four to six arrays
    """
    X_train, X_test, y_train, y_test = train_test_split(
        df.drop(label_name, axis=1).values,
        df[label_name].values,
        test_size=test_size,
        random_state=random_state,
        stratify=df[label_name].values,
    )
    arrlist = [X_test, y_test]
    if valid_size is not None:
        X_train, X_valid, y_train, y_valid = train_test_split(
            X_train,
            y_train,
            test_size=valid_size,
            random_state=random_state,
            stratify=y_train,
        )
        arrlist = [X_valid, y_valid] + arrlist
    X_over, y_over = RandomOverSampler().fit_resample(X_train, y_train)
    vc = np.unique(y_over, return_counts=True)
    print(
        "These labels \t\t" + str(vc[0]) + "\nare oversampled to " + str(vc[1])
    )
    arrlist = [X_over, y_over] + arrlist
    return arrlist
